from typing import Any, Callable, Literal, Optional, Self, TypeVarTuple, overload

import numpy as np
from numpy.typing import NDArray
from scipy.optimize import newton
from scipy.special import digamma, exp1, gamma, gammaincc, gammainccinv, polygamma
from typing_extensions import override

from relife.lifetime_model import LifetimeRegression
from relife.quadrature import laguerre_quadrature, legendre_quadrature

from ..data import LifetimeData
from ..likelihood import LikelihoodFromLifetimes
from ._base import LifetimeDistribution, ParametricLifetimeModel


class Exponential(LifetimeDistribution):
    # noinspection PyUnresolvedReferences
    r"""
    Exponential lifetime distribution.

    The exponential distribution is a 1-parameter distribution with
    :math:`(\lambda)`. The probability density function is:

    .. math::

        f(t) = \lambda e^{-\lambda t}

    where:
        - :math:`\lambda > 0`, the rate parameter,
        - :math:`t\geq 0`, the operating time, age, cycles, etc.

    |

    Parameters
    ----------
    rate : float, optional
        rate parameter

    Attributes
    ----------
    fitting_results : FittingResults, default is None
        An object containing fitting results (AIC, BIC, etc.). If the model is not fitted, the value is None.
    nb_params
    params
    params_names
    plot
    rate
    """

    def __init__(self, rate: Optional[float] = None):
        super().__init__(rate=rate)

    @property
    def rate(self) -> float:  # optional but better for clarity and type checking
        """Get the current rate value.

        Returns
        -------
        float
        """
        return self._params.get_param_value("rate")

    def hf(self, time: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        return self.rate * np.ones_like(time)

    def chf(self, time: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        return np.asarray(self.rate) * time

    @override
    def mean(self) -> np.float64:
        return 1 / np.asarray(self.rate)

    @override
    def var(self) -> np.float64:
        return 1 / np.asarray(self.rate) ** 2

    @override
    def mrl(self, time: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        return 1 / self.rate * np.ones_like(time)

    @override
    def ichf(self, cumulative_hazard_rate: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        return cumulative_hazard_rate / np.asarray(self.rate)

    @overload
    def jac_hf(
        self,
        time: float | NDArray[np.float64],
        *,
        asarray: Literal[False],
    ) -> tuple[np.float64 | NDArray[np.float64], ...]: ...

    @overload
    def jac_hf(
        self,
        time: float | NDArray[np.float64],
        *,
        asarray: Literal[True],
    ) -> np.float64 | NDArray[np.float64]: ...

    def jac_hf(self, time: float | NDArray[np.float64], *, asarray: bool = False) -> np.float64 | NDArray[np.float64]:
        if isinstance(time, np.ndarray):
            jac = np.expand_dims(np.ones_like(time), axis=0).copy()
        else:
            jac = np.array([1], dtype=np.float64)
        if not asarray:
            return np.unstack(jac)
        return jac

    @overload
    def jac_chf(
        self,
        time: float | NDArray[np.float64],
        *,
        asarray: Literal[False],
    ) -> tuple[np.float64 | NDArray[np.float64], ...]: ...

    @overload
    def jac_chf(
        self,
        time: float | NDArray[np.float64],
        *,
        asarray: Literal[True],
    ) -> np.float64 | NDArray[np.float64]: ...

    def jac_chf(
        self, time: float | NDArray[np.float64], *, asarray: bool = False
    ) -> np.float64 | NDArray[np.float64] | tuple[np.float64 | NDArray[np.float64], ...]:
        if isinstance(time, np.ndarray):
            jac = np.expand_dims(time, axis=0).copy()
        else:
            jac = np.array([time], dtype=np.float64)
        if not asarray:
            return np.unstack(jac)
        return jac

    def dhf(self, time: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        if isinstance(time, np.ndarray):
            return np.zeros_like(time)
        return np.asarray(0.0)


class Weibull(LifetimeDistribution):
    # noinspection PyUnresolvedReferences
    r"""
    Weibull lifetime distribution.

    The Weibull distribution is a 2-parameter distribution with
    :math:`(c,\lambda)`. The probability density function is:

    .. math::

        f(t) = c \lambda (\lambda t)^{c-1} e^{-(\lambda t)^c}

    where:
        - :math:`c > 0`, the shape parameter,
        - :math:`\lambda > 0`, the rate parameter,
        - :math:`t\geq 0`, the operating time, age, cycles, etc.

    Parameters
    ----------
    shape : float, optional
        shape parameter
    rate : float, optional
        rate parameter

    Attributes
    ----------
    fitting_results : FittingResults, default is None
        An object containing fitting results (AIC, BIC, etc.). If the model is not fitted, the value is None.
    nb_params
    params
    params_names
    plot
    shape
    rate
    """

    def __init__(self, shape: Optional[float] = None, rate: Optional[float] = None):
        super().__init__(shape=shape, rate=rate)

    @property
    def shape(self) -> float:  # optional but better for clarity and type checking
        """Get the current shape value.

        Returns
        -------
        float
        """
        return self._params.get_param_value("shape")

    @property
    def rate(self) -> float:  # optional but better for clarity and type checking
        """Get the current rate value.

        Returns
        -------
        float
        """
        return self._params.get_param_value("rate")

    def hf(self, time: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        return self.shape * self.rate * (self.rate * np.asarray(time)) ** (self.shape - 1)

    def chf(self, time: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        return (self.rate * np.asarray(time)) ** self.shape

    @override
    def mean(self) -> np.float64:
        return gamma(1 + 1 / self.shape) / self.rate

    @override
    def var(self) -> np.float64:
        return gamma(1 + 2 / self.shape) / self.rate**2 - self.mean() ** 2

    @override
    def mrl(self, time: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        return (
            gamma(1 / self.shape)
            / (self.rate * self.shape * self.sf(time))
            * gammaincc(
                1 / self.shape,
                (self.rate * time) ** self.shape,
            )
        )

    @override
    def ichf(self, cumulative_hazard_rate: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        return np.asarray(cumulative_hazard_rate) ** (1 / self.shape) / self.rate

    @overload
    def jac_hf(
        self,
        time: float | NDArray[np.float64],
        *,
        asarray: Literal[False],
    ) -> tuple[np.float64 | NDArray[np.float64], ...]: ...

    @overload
    def jac_hf(
        self,
        time: float | NDArray[np.float64],
        *,
        asarray: Literal[True],
    ) -> np.float64 | NDArray[np.float64]: ...

    def jac_hf(
        self, time: float | NDArray[np.float64], *, asarray: bool = False
    ) -> tuple[np.float64, np.float64] | tuple[NDArray[np.float64], NDArray[np.float64]]:
        jac = (
            self.rate * (self.rate * time) ** (self.shape - 1) * (1 + self.shape * np.log(self.rate * time)),
            self.shape**2 * (self.rate * time) ** (self.shape - 1),
        )
        if asarray:
            return np.stack(jac)
        return jac

    @overload
    def jac_chf(
        self,
        time: float | NDArray[np.float64],
        *,
        asarray: Literal[False],
    ) -> tuple[np.float64 | NDArray[np.float64], ...]: ...

    @overload
    def jac_chf(
        self,
        time: float | NDArray[np.float64],
        *,
        asarray: Literal[True],
    ) -> np.float64 | NDArray[np.float64]: ...

    def jac_chf(
        self, time: float | NDArray[np.float64], *, asarray: bool = False
    ) -> tuple[np.float64, np.float64] | tuple[NDArray[np.float64], NDArray[np.float64]]:
        jac = (
            np.log(self.rate * time) * (self.rate * time) ** self.shape,
            self.shape * time * (self.rate * time) ** (self.shape - 1),
        )
        if asarray:
            return np.stack(jac)
        return jac

    def dhf(self, time: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        time = np.asarray(time)
        return self.shape * (self.shape - 1) * self.rate**2 * (self.rate * time) ** (self.shape - 2)


class Gompertz(LifetimeDistribution):
    # noinspection PyUnresolvedReferences
    r"""
    Gompertz lifetime distribution.

    The Gompertz distribution is a 2-parameter distribution with
    :math:`(c,\lambda)`. The probability density function is:

    .. math::

        f(t) = c \lambda e^{\lambda t} e^{ -c \left( e^{\lambda t}-1 \right) }

    where:

        - :math:`c > 0`, the shape parameter,
        - :math:`\lambda > 0`, the rate parameter,
        - :math:`t\geq 0`, the operating time, age, cycles, etc.

    |

    Parameters
    ----------
    shape : float, optional
        shape parameter
    rate : float, optional
        rate parameter

    Attributes
    ----------
    fitting_results : FittingResults, default is None
        An object containing fitting results (AIC, BIC, etc.). If the model is not fitted, the value is None.
    nb_params
    params
    params_names
    plot
    shape
    rate
    """

    def __init__(self, shape: Optional[float] = None, rate: Optional[float] = None):
        super().__init__(shape=shape, rate=rate)

    @property
    def shape(self) -> float:  # optional but better for clarity and type checking
        """Get the current shape value.

        Returns
        -------
        float
        """
        return self._params.get_param_value("shape")

    @property
    def rate(self) -> float:  # optional but better for clarity and type checking
        """Get the current rate value.

        Returns
        -------
        float
        """
        return self._params.get_param_value("rate")

    def hf(self, time: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        return self.shape * self.rate * np.exp(self.rate * time)

    def chf(self, time: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        return self.shape * np.expm1(self.rate * time)

    @override
    def mean(self) -> np.float64:
        return np.exp(self.shape) * exp1(self.shape) / self.rate

    @override
    def var(self) -> np.float64:
        return polygamma(1, 1) / self.rate**2

    @override
    def mrl(self, time: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        z = self.shape * np.exp(self.rate * time)
        return np.exp(z) * exp1(z) / self.rate

    @override
    def ichf(self, cumulative_hazard_rate: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        return 1 / self.rate * np.log1p(cumulative_hazard_rate / self.shape)

    @overload
    def jac_hf(
        self,
        time: float | NDArray[np.float64],
        *,
        asarray: Literal[False],
    ) -> tuple[np.float64 | NDArray[np.float64], ...]: ...

    @overload
    def jac_hf(
        self,
        time: float | NDArray[np.float64],
        *,
        asarray: Literal[True],
    ) -> np.float64 | NDArray[np.float64]: ...

    def jac_hf(
        self, time: float | NDArray[np.float64], *, asarray: bool = False
    ) -> tuple[np.float64, np.float64] | tuple[NDArray[np.float64], NDArray[np.float64]]:
        jac = (
            self.rate * np.exp(self.rate * time),
            self.shape * np.exp(self.rate * time) * (1 + self.rate * time),
        )
        if asarray:
            return np.stack(jac)
        return jac

    @overload
    def jac_chf(
        self,
        time: float | NDArray[np.float64],
        *,
        asarray: Literal[False],
    ) -> tuple[np.float64 | NDArray[np.float64], ...]: ...

    @overload
    def jac_chf(
        self,
        time: float | NDArray[np.float64],
        *,
        asarray: Literal[True],
    ) -> np.float64 | NDArray[np.float64]: ...

    def jac_chf(
        self, time: float | NDArray[np.float64], *, asarray: bool = False
    ) -> tuple[np.float64, np.float64] | tuple[NDArray[np.float64], NDArray[np.float64]]:
        jac = (
            np.expm1(self.rate * time),
            self.shape * time * np.exp(self.rate * time),
        )
        if asarray:
            return np.stack(jac)
        return jac

    def dhf(self, time: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        return self.shape * self.rate**2 * np.exp(self.rate * time)


class Gamma(LifetimeDistribution):
    # noinspection PyUnresolvedReferences
    r"""
    Gamma lifetime distribution.

    The Gamma distribution is a 2-parameter distribution with
    :math:`(c,\lambda)`. The probability density function is:

    .. math::

        f(t) = \frac{\lambda^c t^{c-1} e^{-\lambda t}}{\Gamma(c)}

    where:

        - :math:`c > 0`, the shape parameter,
        - :math:`\lambda > 0`, the rate parameter,
        - :math:`t\geq 0`, the operating time, age, cycles, etc.

    |

    Parameters
    ----------
    shape : float, optional
        shape parameter
    rate : float, optional
        rate parameter

    Attributes
    ----------
    fitting_results : FittingResults, default is None
        An object containing fitting results (AIC, BIC, etc.). If the model is not fitted, the value is None.
    nb_params
    params
    params_names
    plot
    shape
    rate
    """

    def __init__(self, shape: Optional[float] = None, rate: Optional[float] = None):
        super().__init__(shape=shape, rate=rate)

    @property
    def shape(self) -> float:  # optional but better for clarity and type checking
        """Get the current shape value.

        Returns
        -------
        float
        """
        return self._params.get_param_value("shape")

    @property
    def rate(self) -> float:  # optional but better for clarity and type checking
        """Get the current rate value.

        Returns
        -------
        float
        """
        return self._params.get_param_value("rate")

    def hf(self, time: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        x = self.rate * time
        return self.rate * x ** (self.shape - 1) * np.exp(-x) / self._uppergamma(x)

    def chf(self, time: float | NDArray[np.float64]) -> float | NDArray[np.float64]:
        x = self.rate * time
        return np.log(gamma(self.shape)) - np.log(self._uppergamma(x))

    @override
    def mean(self) -> np.float64:
        return np.asarray(self.shape / self.rate)

    @override
    def var(self) -> np.float64:
        return np.asarray(self.shape / (self.rate**2))

    @override
    def ichf(self, cumulative_hazard_rate: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        return 1 / self.rate * gammainccinv(self.shape, np.exp(-cumulative_hazard_rate))

    def _uppergamma(self, x: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        return gammaincc(self.shape, x) * gamma(self.shape)

    def _jac_uppergamma_shape(self, x: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        return laguerre_quadrature(lambda s: np.log(s) * s ** (self.shape - 1), x, deg=100)

    @overload
    def jac_hf(
        self,
        time: float | NDArray[np.float64],
        *,
        asarray: Literal[False],
    ) -> tuple[np.float64 | NDArray[np.float64], ...]: ...

    @overload
    def jac_hf(
        self,
        time: float | NDArray[np.float64],
        *,
        asarray: Literal[True],
    ) -> np.float64 | NDArray[np.float64]: ...

    def jac_hf(
        self, time: float | NDArray[np.float64], *, asarray: bool = False
    ) -> tuple[np.float64, np.float64] | tuple[NDArray[np.float64], NDArray[np.float64]]:
        x = self.rate * time
        y = x ** (self.shape - 1) * np.exp(-x) / self._uppergamma(x) ** 2
        jac = (
            y * ((self.rate * np.log(x) * self._uppergamma(x)) - self.rate * self._jac_uppergamma_shape(x)),
            y * ((self.shape - x) * self._uppergamma(x) + x**self.shape * np.exp(-x)),
        )
        if asarray:
            return np.stack(jac)
        return jac

    @overload
    def jac_chf(
        self,
        time: float | NDArray[np.float64],
        *,
        asarray: Literal[False],
    ) -> tuple[np.float64 | NDArray[np.float64], ...]: ...

    @overload
    def jac_chf(
        self,
        time: float | NDArray[np.float64],
        *,
        asarray: Literal[True],
    ) -> np.float64 | NDArray[np.float64]: ...

    def jac_chf(
        self, time: float | NDArray[np.float64], *, asarray: bool = False
    ) -> tuple[np.float64, np.float64] | tuple[NDArray[np.float64], NDArray[np.float64]]:
        x = self.rate * time
        jac = (
            digamma(self.shape) - self._jac_uppergamma_shape(x) / self._uppergamma(x),
            (x ** (self.shape - 1) * time * np.exp(-x) / self._uppergamma(x)),
        )
        if asarray:
            return np.stack(jac)
        return jac

    def dhf(self, time: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        return self.hf(time) * ((self.shape - 1) / time - self.rate + self.hf(time))

    @override
    def mrl(self, time: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        return super().mrl(time)


class LogLogistic(LifetimeDistribution):
    # noinspection PyUnresolvedReferences
    r"""
    Log-logistic probability distribution.

    The Log-logistic distribution is defined as a 2-parameter distribution
    :math:`(c, \lambda)`. The probability density function is:

    .. math::

        f(t) = \frac{c \lambda^c t^{c-1}}{(1+(\lambda t)^{c})^2}

    where:

        - :math:`c > 0`, the shape parameter,
        - :math:`\lambda > 0`, the rate parameter,
        - :math:`t\geq 0`, the operating time, age, cycles, etc.

    |

    Parameters
    ----------
    shape : float, optional
        shape parameter
    rate : float, optional
        rate parameter

    Attributes
    ----------
    fitting_results : FittingResults, default is None
        An object containing fitting results (AIC, BIC, etc.). If the model is not fitted, the value is None.
    nb_params
    params
    params_names
    plot
    shape
    rate
    """

    def __init__(self, shape: Optional[float] = None, rate: Optional[float] = None):
        super().__init__(shape=shape, rate=rate)

    @property
    def shape(self) -> float:  # optional but better for clarity and type checking
        """Get the current shape value.

        Returns
        -------
        float
        """
        return self._params.get_param_value("shape")

    @property
    def rate(self) -> float:  # optional but better for clarity and type checking
        """Get the current rate value.

        Returns
        -------
        float
        """
        return self._params.get_param_value("rate")

    def hf(self, time: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        x = self.rate * np.asarray(time)
        return self.shape * self.rate * x ** (self.shape - 1) / (1 + x**self.shape)

    def chf(self, time: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        x = self.rate * time
        return np.log(1 + x**self.shape)

    @override
    def mean(self) -> np.float64:
        b = np.pi / self.shape
        if self.shape <= 1:
            raise ValueError(f"Expectancy only defined for shape > 1: shape = {self.shape}")
        return b / (self.rate * np.sin(b))

    @override
    def var(self) -> np.float64:
        b = np.pi / self.shape
        if self.shape <= 2:
            raise ValueError(f"Variance only defined for shape > 2: shape = {self.shape}")
        return (1 / self.rate**2) * (2 * b / np.sin(2 * b) - b**2 / (np.sin(b) ** 2))

    @override
    def ichf(self, cumulative_hazard_rate: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        return ((np.exp(cumulative_hazard_rate) - 1) ** (1 / self.shape)) / self.rate

    @overload
    def jac_hf(
        self,
        time: float | NDArray[np.float64],
        *,
        asarray: Literal[False],
    ) -> tuple[np.float64 | NDArray[np.float64], ...]: ...

    @overload
    def jac_hf(
        self,
        time: float | NDArray[np.float64],
        *,
        asarray: Literal[True],
    ) -> np.float64 | NDArray[np.float64]: ...

    def jac_hf(
        self, time: float | NDArray[np.float64], *, asarray: bool = False
    ) -> tuple[np.float64, np.float64] | tuple[NDArray[np.float64], NDArray[np.float64]]:
        x = self.rate * time
        jac = (
            (self.rate * x ** (self.shape - 1) / (1 + x**self.shape) ** 2)
            * (1 + x**self.shape + self.shape * np.log(self.rate * time)),
            (self.rate * x ** (self.shape - 1) / (1 + x**self.shape) ** 2) * (self.shape**2 / self.rate),
        )
        if asarray:
            return np.stack(jac)
        return jac

    @overload
    def jac_chf(
        self,
        time: float | NDArray[np.float64],
        *,
        asarray: Literal[False],
    ) -> tuple[np.float64 | NDArray[np.float64], ...]: ...

    @overload
    def jac_chf(
        self,
        time: float | NDArray[np.float64],
        *,
        asarray: Literal[True],
    ) -> np.float64 | NDArray[np.float64]: ...

    def jac_chf(
        self, time: float | NDArray[np.float64], *, asarray: bool = False
    ) -> tuple[np.float64, np.float64] | tuple[NDArray[np.float64], NDArray[np.float64]]:
        x = self.rate * time
        jac = (
            (x**self.shape / (1 + x**self.shape)) * np.log(self.rate * time),
            (x**self.shape / (1 + x**self.shape)) * (self.shape / self.rate),
        )
        if asarray:
            return np.stack(jac)
        return jac

    def dhf(self, time: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        x = self.rate * np.asarray(time)
        return (
            self.shape
            * self.rate**2
            * x ** (self.shape - 2)
            * (self.shape - 1 - x**self.shape)
            / (1 + x**self.shape) ** 2
        )

    @override
    def mrl(self, time: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        return super().mrl(time)


Args = TypeVarTuple("Args")


class EquilibriumDistribution(ParametricLifetimeModel[*tuple[float | NDArray[np.float64], ...]]):
    r"""Equilibrium distribution.

    The equilibirum distribution is the distrbution computed from a lifetime
    core that makes the associated delayed renewal stochastic_process stationary.

    Parameters
    ----------
    baseline : BaseLifetimeModel
        Underlying lifetime core.

    References
    ----------
    .. [1] Ross, S. M. (1996). Stochastic stochastic_process. New York: Wiley.
    """

    # can't expect baseline to be FrozenParametricLifetimeModel too because it does not have freeze_args
    def __init__(self, baseline: ParametricLifetimeModel[*tuple[float | NDArray[np.float64], ...]]):
        super().__init__()
        self.baseline = baseline

    @property
    def args_names(self) -> tuple[str, ...]:
        return self.baseline.args_names

    @override
    def cdf(
        self, time: float | NDArray[np.float64], *args: float | NDArray[np.float64]
    ) -> np.float64 | NDArray[np.float64]:
        return legendre_quadrature(lambda x: self.baseline.sf(x, *args), 0, time) / self.baseline.mean(*args)

    def sf(
        self, time: float | NDArray[np.float64], *args: float | NDArray[np.float64]
    ) -> np.float64 | NDArray[np.float64]:
        return 1 - self.cdf(time, *args)

    def pdf(
        self, time: float | NDArray[np.float64], *args: float | NDArray[np.float64]
    ) -> np.float64 | NDArray[np.float64]:
        return self.baseline.sf(time, *args) / self.baseline.mean(*args)

    def hf(
        self, time: float | NDArray[np.float64], *args: float | NDArray[np.float64]
    ) -> np.float64 | NDArray[np.float64]:
        return 1 / self.baseline.mrl(time, *args)

    def chf(
        self, time: float | NDArray[np.float64], *args: float | NDArray[np.float64]
    ) -> np.float64 | NDArray[np.float64]:
        return -np.log(self.sf(time, *args))

    @override
    def isf(
        self, probability: float | NDArray[np.float64], *args: float | NDArray[np.float64]
    ) -> np.float64 | NDArray[np.float64]:
        return newton(
            lambda x: self.sf(x, *args) - probability,
            self.baseline.isf(probability, *args),
            args=args,
        )

    @override
    def ichf(
        self,
        cumulative_hazard_rate: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
    ) -> np.float64 | NDArray[np.float64]:
        return self.isf(np.exp(-cumulative_hazard_rate), *args)


class MinimumDistribution(ParametricLifetimeModel[int | NDArray[np.int64], *tuple[float | NDArray[np.float64], ...]]):
    r"""Series structure of n identical and independent components.

    The hazard function of the system is given by:

    .. math::

        h(t) = n \cdot  h_0(t)

    where :math:`h_0` is the baseline hazard function of the components.

    Examples
    --------

    Computing the survival (or reliability) function for 3 structures of 3,6 and
    9 identical and idependent components:

    .. code-block::

        model = MinimumDistribution(Weibull(2, 0.05))
        t = np.arange(0, 10, 0.1)
        n = np.array([3, 6, 9]).reshape(-1, 1)
        model.sf(t, n)
    """

    def __init__(self, baseline: LifetimeDistribution | LifetimeRegression):
        super().__init__()
        self.baseline = baseline

    @override
    def sf(
        self, time: float | NDArray[np.float64], n: int | NDArray[np.int64], *args: float | NDArray[np.float64]
    ) -> np.float64 | NDArray[np.float64]:
        return super().sf(time)

    @override
    def pdf(
        self, time: float | NDArray[np.float64], n: int | NDArray[np.int64], *args: float | NDArray[np.float64]
    ) -> np.float64 | NDArray[np.float64]:
        return super().pdf(time)

    @override
    def hf(
        self, time: float | NDArray[np.float64], n: int | NDArray[np.int64], *args: float | NDArray[np.float64]
    ) -> np.float64 | NDArray[np.float64]:
        return n * self.baseline.hf(time, *args)

    @override
    def chf(
        self, time: float | NDArray[np.float64], n: int | NDArray[np.int64], *args: float | NDArray[np.float64]
    ) -> np.float64 | NDArray[np.float64]:
        return n * self.baseline.chf(time, *args)

    @override
    def ichf(
        self,
        cumulative_hazard_rate: float | NDArray[np.float64],
        n: int | NDArray[np.int64],
        *args: float | NDArray[np.float64],
    ) -> np.float64 | NDArray[np.float64]:
        return self.baseline.ichf(cumulative_hazard_rate / n, *args)

    def dhf(
        self, time: float | NDArray[np.float64], n: int | NDArray[np.int64], *args: float | NDArray[np.float64]
    ) -> np.float64 | NDArray[np.float64]:
        return n * self.baseline.dhf(time, *args)

    def jac_chf(
        self,
        time: float | NDArray[np.float64],
        n: int | NDArray[np.int64],
        *args: float | NDArray[np.float64],
        asarray: bool = False,
    ) -> np.float64 | NDArray[np.float64] | tuple[np.float64 | NDArray[np.float64], ...]:
        return n * self.baseline.jac_chf(time, *args, asarray=asarray)

    def jac_hf(
        self,
        time: float | NDArray[np.float64],
        n: int | NDArray[np.int64],
        *args: float | NDArray[np.float64],
        asarray: bool = False,
    ) -> np.float64 | NDArray[np.float64] | tuple[np.float64 | NDArray[np.float64], ...]:
        return n * self.baseline.jac_hf(time, *args, asarray=asarray)

    def jac_sf(
        self,
        time: float | NDArray[np.float64],
        n: int | NDArray[np.int64],
        *args: float | NDArray[np.float64],
        asarray: bool = False,
    ) -> np.float64 | NDArray[np.float64] | tuple[np.float64 | NDArray[np.float64], ...]:
        jac_chf, sf = self.jac_chf(time, n, *args, asarray=True), self.sf(time)
        jac = -jac_chf * sf
        if not asarray:
            return np.unstack(jac)
        return jac

    def jac_cdf(
        self,
        time: float | NDArray[np.float64],
        n: int | NDArray[np.int64],
        *args: float | NDArray[np.float64],
        asarray: bool = False,
    ) -> np.float64 | NDArray[np.float64] | tuple[np.float64 | NDArray[np.float64], ...]:
        jac = -self.jac_sf(time, n, *args, asarray=True)
        if not asarray:
            return np.unstack(jac)
        return jac

    def jac_pdf(
        self,
        time: float | NDArray[np.float64],
        n: int | NDArray[np.int64],
        *args: float | NDArray[np.float64],
        asarray: bool = False,
    ) -> np.float64 | NDArray[np.float64] | tuple[np.float64 | NDArray[np.float64], ...]:
        jac_hf, hf = self.jac_hf(time, n, *args, asarray=True), self.hf(time)
        jac_sf, sf = self.jac_sf(time, n, *args, asarray=True), self.sf(time)
        jac = jac_hf * sf + jac_sf * hf
        if not asarray:
            return np.unstack(jac)
        return jac

    @override
    def ls_integrate(
        self,
        func: Callable[[float | NDArray[np.float64]], np.float64 | NDArray[np.float64]],
        a: float | NDArray[np.float64],
        b: float | NDArray[np.float64],
        n: int | NDArray[np.int64],
        *args: float | NDArray[np.float64],
        deg: int = 10,
    ) -> np.float64 | NDArray[np.float64]:
        return super().ls_integrate(func, a, b, n, *args, deg=deg)

    def fit(
        self,
        time: NDArray[np.float64],
        n: NDArray[np.int64],
        *args: NDArray[np.float64],
        event: Optional[NDArray[np.bool_]] = None,
        entry: Optional[NDArray[np.float64]] = None,
        departure: Optional[NDArray[np.float64]] = None,
        **kwargs: Any,
    ) -> Self:
        # initialize params structure (number of parameters in params tree)
        if isinstance(self.baseline, LifetimeRegression):
            self.baseline._init_coefficients(args[0], *args[1:])
        lifetime_data: LifetimeData = LifetimeData(time, event=event, entry=entry, departure=departure, args=(n, *args))
        likelihood = LikelihoodFromLifetimes(self, lifetime_data)
        fitting_results = likelihood.maximum_likelihood_estimation(**kwargs)
        self.params = fitting_results.optimal_params
        self.fitting_results = fitting_results
        return self


TIME_BASE_DOCSTRING = """
{name}.

Parameters
----------
time : float or np.ndarray
    Elapsed time value(s) at which to compute the function.
    If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

Returns
-------
np.float64 or np.ndarray
    Function values at each given time(s).
"""


JAC_BASE_DOCSTRING = """
{name}.

Parameters
----------
time : float or np.ndarray
    Elapsed time value(s) at which to compute the function.
    If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
asarray : bool, default is False

Returns
-------
np.float64, np.ndarray or tuple of np.float64 or np.ndarray
    The derivatives with respect to each parameter. If ``asarray`` is False, the function returns a tuple containing
    the same number of elements as parameters. If ``asarray`` is True, the function returns an ndarray
    whose first dimension equals the number of parameters. This output is equivalent to applying ``np.stack`` on the output
    tuple when ``asarray`` is False.
"""


MOMENT_BASE_DOCSTRING = """
{name}.

Returns
-------
np.float64
    {name} value.
"""


PROBABILITY_BASE_DOCSTRING = """
{name}.

Parameters
----------
probability : float or np.ndarray
    Probability value(s) at which to compute the function.
    If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

Returns
-------
np.float64 or np.ndarray
    Function values at each given probability value(s).
"""


for class_obj in (Exponential, Weibull, Gompertz, Gamma, LogLogistic):
    class_obj.sf.__doc__ = TIME_BASE_DOCSTRING.format(name="The survival function")
    class_obj.hf.__doc__ = TIME_BASE_DOCSTRING.format(name="The hazard function")
    class_obj.chf.__doc__ = TIME_BASE_DOCSTRING.format(name="The cumulative hazard function")
    class_obj.pdf.__doc__ = TIME_BASE_DOCSTRING.format(name="The probability density function")
    class_obj.cdf.__doc__ = TIME_BASE_DOCSTRING.format(name="The cumulative distribution function")
    class_obj.mrl.__doc__ = TIME_BASE_DOCSTRING.format(name="The mean residual life function")
    class_obj.dhf.__doc__ = TIME_BASE_DOCSTRING.format(name="The derivative of the hazard function")
    class_obj.jac_hf.__doc__ = JAC_BASE_DOCSTRING.format(name="The jacobian of the hazard function")
    class_obj.jac_chf.__doc__ = JAC_BASE_DOCSTRING.format(name="The jacobian of the cumulative hazard function")
    class_obj.jac_sf.__doc__ = JAC_BASE_DOCSTRING.format(name="The jacobian of the survival function")
    class_obj.jac_pdf.__doc__ = JAC_BASE_DOCSTRING.format(name="The jacobian of the probability density function")
    class_obj.jac_cdf.__doc__ = JAC_BASE_DOCSTRING.format(name="The jacobian of the cumulative distribution function")

    class_obj.ppf.__doc__ = PROBABILITY_BASE_DOCSTRING.format(name="The percent point function")
    class_obj.ppf.__doc__ += f"""
    Notes
    -----
    The ``ppf`` is the inverse of :py:meth:`~{class_obj}.cdf`.
    """
    class_obj.isf.__doc__ = PROBABILITY_BASE_DOCSTRING.format(name="Inverse survival function")

    class_obj.rvs.__doc__ = """
    Random variable sampling.

    Parameters
    ----------
    size : int, (int,) or (int, int)
        Size of the generated sample. If size is ``n`` or ``(n,)``, n samples are generated. If size is ``(m,n)``, a 
        2d array of samples is generated. 
    return_event : bool, default is False
        If True, returns event indicators along with the sample time values.
    random_entry : bool, default is False
        If True, returns corresponding entry values of the sample time values.
    seed : optional int, default is None
        Random seed used to fix random sampling.

    Returns
    -------
    float, ndarray or tuple of float or ndarray
        The sample values. If either ``return_event`` or ``random_entry`` is True, returns a tuple containing
        the time values followed by event values, entry values or both.
    """

    class_obj.plot.__doc__ = """
    Provides access to plotting functionality for this distribution.
    """

    class_obj.ls_integrate.__doc__ = """
    Lebesgue-Stieltjes integration.

    Parameters
    ----------
    func : callable (in : 1 ndarray , out : 1 ndarray)
        The callable must have only one ndarray object as argument and one ndarray object as output
    a : ndarray (maximum number of dimension is 2)
        Lower bound(s) of integration.
    b : ndarray (maximum number of dimension is 2)
        Upper bound(s) of integration. If lower bound(s) is infinite, use np.inf as value.)
    deg : int, default 10
        Degree of the polynomials interpolation

    Returns
    -------
    np.ndarray
        Lebesgue-Stieltjes integral of func from `a` to `b`.
    """

    class_obj.moment.__doc__ = """
    n-th order moment

    Parameters
    ----------
    n : order of the moment, at least 1.

    Returns
    -------
    np.float64
        n-th order moment.
    """
    class_obj.mean.__doc__ = MOMENT_BASE_DOCSTRING.format(name="The mean")
    class_obj.var.__doc__ = MOMENT_BASE_DOCSTRING.format(name="The variance")
    class_obj.median.__doc__ = MOMENT_BASE_DOCSTRING.format(name="The median")

    class_obj.ichf.__doc__ = """
    Inverse cumulative hazard function.
    
    Parameters
    ----------
    cumulative_hazard_rate : float or np.ndarray
        Cumulative hazard rate value(s) at which to compute the function.
        If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
    
    Returns
    -------
    np.float64 or np.ndarray
        Function values at each given cumulative hazard rate(s).
    """

    class_obj.fit.__doc__ = """
    Estimation of parameters.

    Parameters
    ----------
    time : ndarray (1d or 2d)
        Observed lifetime values.
    event : ndarray of boolean values (1d), default is None
        Boolean indicators tagging lifetime values as right censored or complete.
    entry : ndarray of float (1d), default is None
        Left truncations applied to lifetime values.
    departure : ndarray of float (1d), default is None
        Right truncations applied to lifetime values.
    **kwargs
        Extra arguments used by `scipy.minimize`. Default values are:
            - `method` : `"L-BFGS-B"`
            - `contraints` : `()`
            - `tol` : `None`
            - `callback` : `None`
            - `options` : `None`
            - `bounds` : `self.params_bounds`
            - `x0` : `self.init_params`

    Returns
    -------
    Self
        The current object with the estimated parameters setted inplace.

    Notes
    -----
    Supported lifetime observations format is either 1d-array or 2d-array. 2d-array is more advanced
    format that allows to pass other information as left-censored or interval-censored values. In this case,
    `event` is not needed as 2d-array encodes right-censored values by itself.
    """



from ._base import BaseAgeReplacementPolicy, BaseOneCycleAgeReplacementPolicy
from .age_replacement_policy import AgeReplacementPolicy, OneCycleAgeReplacementPolicy
from .run_to_failure_policy import OneCycleRunToFailurePolicy, RunToFailurePolicy



from __future__ import annotations

from typing import Callable, Literal, Optional, Union, overload

import numpy as np
from numpy.typing import NDArray
from typing_extensions import override

from ._base import FrozenParametricLifetimeModel, ParametricLifetimeModel


def reshape_ar_or_a0(name: str, value: float | NDArray[np.float64]) -> NDArray[np.float64]:
    value = np.asarray(value)  # in shape : (), (m,) or (m, 1)
    if value.ndim > 2 or (value.ndim == 2 and value.shape[-1] != 1):
        raise ValueError(f"Incorrect {name} shape. Got {value.shape}. Expected (), (m,) or (m, 1)")
    if value.ndim == 1:
        value = value.reshape(-1, 1)
    return value  # out shape: () or (m, 1)


# note that AgeReplacementModel does not preserve generic : at the moment, additional args are supposed to be always float | NDArray[np.float64]
class AgeReplacementModel(
    ParametricLifetimeModel[float | NDArray[np.float64], *tuple[float | NDArray[np.float64], ...]]
):
    # noinspection PyUnresolvedReferences
    r"""
    Age replacement model.

    Lifetime model where the assets are replaced at age :math:`a_r`. This is equivalent to the model of :math:`\min(X,a_r)` where
    :math:`X` is a baseline lifetime and :math:`a_r` is the age of replacement.

    Parameters
    ----------
    baseline : any parametric lifetime model (frozen lifetime model works)
        The base lifetime model without conditional probabilities

    Attributes
    ----------
    baseline
    nb_params
    params
    params_names
    plot
    """

    def __init__(
        self,
        baseline: (
            ParametricLifetimeModel[*tuple[float | NDArray[np.float64], ...]]
            | FrozenParametricLifetimeModel[*tuple[float | NDArray[np.float64], ...]]
        ),
    ):
        super().__init__()
        self.baseline = baseline

    @property
    def args_names(self) -> tuple[str, *tuple[str, ...]]:
        return ("ar",) + self.baseline.args_names

    def sf(
        self,
        time: float | NDArray[np.float64],
        ar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
    ) -> NDArray[np.float64]:
        ar = reshape_ar_or_a0("ar", ar)
        return np.where(time < ar, self.baseline.sf(time, *args), 0.0)

    def hf(
        self,
        time: float | NDArray[np.float64],
        ar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
    ) -> NDArray[np.float64]:
        ar = reshape_ar_or_a0("ar", ar)
        return np.where(time < ar, self.baseline.hf(time, *args), 0.0)

    def chf(
        self,
        time: float | NDArray[np.float64],
        ar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
    ) -> NDArray[np.float64]:
        ar = reshape_ar_or_a0("ar", ar)
        return np.where(time < ar, self.baseline.chf(time, *args), 0.0)

    @override
    def isf(
        self,
        probability: float | NDArray[np.float64],
        ar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
    ) -> NDArray[np.float64]:
        ar = reshape_ar_or_a0("ar", ar)
        return np.minimum(self.baseline.isf(probability, *args), ar)

    @override
    def ichf(
        self,
        cumulative_hazard_rate: NDArray[np.float64],
        ar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
    ) -> NDArray[np.float64]:
        ar = reshape_ar_or_a0("ar", ar)
        return np.minimum(self.baseline.ichf(cumulative_hazard_rate, *args), ar)

    def pdf(
        self,
        time: float | NDArray[np.float64],
        ar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
    ) -> NDArray[np.float64]:
        ar = reshape_ar_or_a0("ar", ar)
        return np.where(time < ar, self.baseline.pdf(time, *args), 0)

    def ppf(
        self,
        probability: float | NDArray[np.float64],
        ar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
    ) -> NDArray[np.float64]:
        ar = reshape_ar_or_a0("ar", ar)
        return self.isf(1 - probability, ar, *args)

    def median(self, ar: float | NDArray[np.float64], *args: float | NDArray[np.float64]) -> NDArray[np.float64]:
        ar = reshape_ar_or_a0("ar", ar)
        return self.ppf(np.array(0.5), ar, *args)

    # def cdf(
    #     self,
    #     time: float | NDArray[np.float64],
    #     ar: float | NDArray[np.float64],
    #     *args: float | NDArray[np.float64],
    # ) -> NDArray[np.float64]:
    #     ar = reshape_ar_or_a0("ar", ar)
    #     return np.where(time < ar, self.baseline.cdf(time, *args), 1.0)

    @override
    def mrl(
        self,
        time: float | NDArray[np.float64],
        ar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
    ) -> NDArray[np.float64]:
        ar = reshape_ar_or_a0("ar", ar)
        ub = np.array(np.inf)
        # ar.shape == (m, 1)
        mask = time >= ar  # (m, 1) or (m, n)
        if np.any(mask):
            time, ub = np.broadcast_arrays(time, ub)
            time = np.ma.MaskedArray(time, mask)  # (m, 1) or (m, n)
            ub = np.ma.MaskedArray(ub, mask)  # (m, 1) or (m, n)
        mu = self.ls_integrate(lambda x: x - time, time, ub, ar, *args, deg=10) / self.sf(
            time, ar, *args
        )  # () or (n,) or (m, n)
        np.ma.filled(mu, 0)
        return np.ma.getdata(mu)

    @overload
    def rvs(
        self,
        size: int | tuple[int] | tuple[int, int],
        ar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        return_event: Literal[False],
        return_entry: Literal[False],
        seed: Optional[int] = None,
    ) -> np.float64 | NDArray[np.float64]: ...

    @overload
    def rvs(
        self,
        size: int | tuple[int] | tuple[int, int],
        ar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        return_event: Literal[True],
        return_entry: Literal[False],
        seed: Optional[int] = None,
    ) -> tuple[np.float64 | NDArray[np.float64], np.bool_ | NDArray[np.bool_]]: ...

    @overload
    def rvs(
        self,
        size: int | tuple[int] | tuple[int, int],
        ar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        return_event: Literal[False],
        return_entry: Literal[True],
        seed: Optional[int] = None,
    ) -> tuple[np.float64 | NDArray[np.float64], np.float64 | NDArray[np.float64]]: ...

    @overload
    def rvs(
        self,
        size: int | tuple[int] | tuple[int, int],
        ar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        return_event: Literal[True],
        return_entry: Literal[True],
        seed: Optional[int] = None,
    ) -> tuple[np.float64 | NDArray[np.float64], np.bool_ | NDArray[np.bool_], np.float64 | NDArray[np.float64]]: ...

    @overload
    def rvs(
        self,
        size: int | tuple[int] | tuple[int, int],
        ar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        return_event: bool = False,
        return_entry: bool = False,
        seed: Optional[int] = None,
    ) -> Union[
        np.float64 | NDArray[np.float64],
        tuple[np.float64 | NDArray[np.float64], np.bool_ | NDArray[np.bool_]],
        tuple[np.float64 | NDArray[np.float64], np.float64 | NDArray[np.float64]],
        tuple[np.float64 | NDArray[np.float64], np.bool_ | NDArray[np.bool_], np.float64 | NDArray[np.float64]],
    ]: ...

    @override
    def rvs(
        self,
        size: int | tuple[int] | tuple[int, int],
        ar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        return_event: bool = False,
        return_entry: bool = False,
        seed: Optional[int] = None,
    ) -> Union[
        np.float64 | NDArray[np.float64],
        tuple[np.float64 | NDArray[np.float64], np.bool_ | NDArray[np.bool_]],
        tuple[np.float64 | NDArray[np.float64], np.float64 | NDArray[np.float64]],
        tuple[np.float64 | NDArray[np.float64], np.bool_ | NDArray[np.bool_], np.float64 | NDArray[np.float64]],
    ]:
        ar = reshape_ar_or_a0("ar", ar)
        baseline_rvs = self.baseline.rvs(size, *args, return_event=return_event, return_entry=return_entry, seed=seed)
        time = baseline_rvs[0] if isinstance(baseline_rvs, tuple) else baseline_rvs
        time = np.minimum(time, ar)  # it may change time shape by broadcasting
        if not return_event and not return_entry:
            return time
        elif return_event and not return_entry:
            event = np.broadcast_to(baseline_rvs[1], time.shape).copy()
            event = np.where(time != ar, event, ~event)
            return time, event
        elif not return_event and return_entry:
            entry = np.broadcast_to(baseline_rvs[1], time.shape).copy()
            return time, entry
        else:
            event, entry = baseline_rvs[1:]
            event = np.broadcast_to(event, time.shape).copy()
            entry = np.broadcast_to(entry, time.shape).copy()
            event = np.where(time != ar, event, ~event)
            return time, event, entry

    @override
    def ls_integrate(
        self,
        func: Callable[[float | np.float64 | NDArray[np.float64]], float | np.float64 | NDArray[np.float64]],
        a: float | NDArray[np.float64],
        b: float | NDArray[np.float64],
        ar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        deg: int = 10,
    ) -> NDArray[np.float64]:
        ar = reshape_ar_or_a0("ar", ar)
        b = np.minimum(ar, b)
        integration = self.baseline.ls_integrate(func, a, b, *args, deg=deg)
        return integration + np.where(b == ar, func(ar) * self.baseline.sf(ar, *args), 0)

    @override
    def moment(
        self, n: int, ar: float | NDArray[np.float64], *args: float | NDArray[np.float64]
    ) -> NDArray[np.float64]:
        ar = reshape_ar_or_a0("ar", ar)
        return self.ls_integrate(
            lambda x: x**n,
            0,
            np.inf,
            ar,
            *args,
            deg=100,
        )

    @override
    def mean(self, ar: float | NDArray[np.float64], *args: float | NDArray[np.float64]) -> NDArray[np.float64]:
        ar = reshape_ar_or_a0("ar", ar)
        return self.moment(1, ar, *args)

    @override
    def var(self, ar: float | NDArray[np.float64], *args: float | NDArray[np.float64]) -> NDArray[np.float64]:
        ar = reshape_ar_or_a0("ar", ar)
        return self.moment(2, ar, *args) - self.moment(1, ar, *args) ** 2


class LeftTruncatedModel(
    ParametricLifetimeModel[float | NDArray[np.float64], *tuple[float | NDArray[np.float64], ...]]
):
    # noinspection PyUnresolvedReferences
    r"""Left truncated model.

    Lifetime model where the assets have already reached the age :math:`a_0`.

    Parameters
    ----------
    baseline : any parametric lifetime model (frozen lifetime model works)
        The base lifetime model without conditional probabilities
    nb_params
    params
    params_names
    plot

    Attributes
    ----------
    baseline
    nb_params
    params
    params_names
    plot
    """

    def __init__(
        self,
        baseline: (
            ParametricLifetimeModel[*tuple[float | NDArray[np.float64], ...]]
            | FrozenParametricLifetimeModel[*tuple[float | NDArray[np.float64], ...]]
        ),
    ):
        super().__init__()
        self.baseline = baseline

    @property
    def args_names(self) -> tuple[str, *tuple[str, ...]]:
        return ("a0",) + self.baseline.args_names

    def sf(
        self,
        time: float | NDArray[np.float64],
        a0: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
    ) -> np.float64 | NDArray[np.float64]:
        a0 = reshape_ar_or_a0("a0", a0)
        return super().sf(time, a0, *args)

    def pdf(
        self,
        time: float | NDArray[np.float64],
        a0: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
    ) -> np.float64 | NDArray[np.float64]:
        a0 = reshape_ar_or_a0("a0", a0)
        return super().pdf(time, a0, *args)

    def isf(
        self,
        probability: NDArray[np.float64],
        a0: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
    ) -> np.float64 | NDArray[np.float64]:
        cumulative_hazard_rate = -np.log(probability + 1e-6)  # avoid division by zero
        a0 = reshape_ar_or_a0("a0", a0)
        return self.ichf(cumulative_hazard_rate, a0, *args)

    def chf(
        self,
        time: float | NDArray[np.float64],
        a0: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
    ) -> np.float64 | NDArray[np.float64]:
        a0 = reshape_ar_or_a0("a0", a0)
        return self.baseline.chf(a0 + time, *args) - self.baseline.chf(a0, *args)

    def cdf(
        self,
        time: float | NDArray[np.float64],
        ar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
    ) -> np.float64 | NDArray[np.float64]:
        ar = reshape_ar_or_a0("ar", ar)
        return super().cdf(time, *(ar, *args))

    def hf(
        self,
        time: float | NDArray[np.float64],
        a0: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
    ) -> np.float64 | NDArray[np.float64]:
        a0 = reshape_ar_or_a0("a0", a0)
        return self.baseline.hf(a0 + time, *args)

    def ichf(
        self,
        cumulative_hazard_rate: float | NDArray[np.float64],
        a0: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
    ) -> np.float64 | NDArray[np.float64]:
        a0 = reshape_ar_or_a0("a0", a0)
        return self.baseline.ichf(cumulative_hazard_rate + self.baseline.chf(a0, *args), *args) - a0

    @overload
    def rvs(
        self,
        size: int | tuple[int] | tuple[int, int],
        a0: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        return_event: Literal[False],
        return_entry: Literal[False],
        seed: Optional[int] = None,
    ) -> np.float64 | NDArray[np.float64]: ...

    @overload
    def rvs(
        self,
        size: int | tuple[int] | tuple[int, int],
        a0: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        return_event: Literal[True],
        return_entry: Literal[False],
        seed: Optional[int] = None,
    ) -> tuple[np.float64 | NDArray[np.float64], np.bool_ | NDArray[np.bool_]]: ...

    @overload
    def rvs(
        self,
        size: int | tuple[int] | tuple[int, int],
        a0: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        return_event: Literal[False],
        return_entry: Literal[True],
        seed: Optional[int] = None,
    ) -> tuple[np.float64 | NDArray[np.float64], np.float64 | NDArray[np.float64]]: ...

    @overload
    def rvs(
        self,
        size: int | tuple[int] | tuple[int, int],
        a0: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        return_event: Literal[True],
        return_entry: Literal[True],
        seed: Optional[int] = None,
    ) -> tuple[np.float64 | NDArray[np.float64], np.bool_ | NDArray[np.bool_], np.float64 | NDArray[np.float64]]: ...

    @overload
    def rvs(
        self,
        size: int | tuple[int] | tuple[int, int],
        a0: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        return_event: bool = False,
        return_entry: bool = False,
        seed: Optional[int] = None,
    ) -> Union[
        np.float64 | NDArray[np.float64],
        tuple[np.float64 | NDArray[np.float64], np.bool_ | NDArray[np.bool_]],
        tuple[np.float64 | NDArray[np.float64], np.float64 | NDArray[np.float64]],
        tuple[np.float64 | NDArray[np.float64], np.bool_ | NDArray[np.bool_], np.float64 | NDArray[np.float64]],
    ]: ...

    @override
    def rvs(
        self,
        size: int | tuple[int] | tuple[int, int],
        a0: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        return_event: bool = False,
        return_entry: bool = False,
        seed: Optional[int] = None,
    ) -> Union[
        np.float64 | NDArray[np.float64],
        tuple[np.float64 | NDArray[np.float64], np.bool_ | NDArray[np.bool_]],
        tuple[np.float64 | NDArray[np.float64], np.float64 | NDArray[np.float64]],
        tuple[np.float64 | NDArray[np.float64], np.bool_ | NDArray[np.bool_], np.float64 | NDArray[np.float64]],
    ]:
        a0 = reshape_ar_or_a0("a0", a0)
        super_rvs = super().rvs(size, *(a0, *args), return_event=return_event, return_entry=return_entry, seed=seed)
        if not return_event and return_entry:
            time, entry = super_rvs
            entry = np.broadcast_to(a0, entry.shape).copy()
            time = time + a0  #  not residual age
            return time, entry
        elif return_event and return_entry:
            time, event, entry = super_rvs
            entry = np.broadcast_to(a0, entry.shape).copy()
            time = time + a0  #  not residual age
            return time, event, entry
        else:
            return super_rvs

    @override
    def ls_integrate(
        self,
        func: Callable[[float | np.float64 | NDArray[np.float64]], float | np.float64 | NDArray[np.float64]],
        a: float | NDArray[np.float64],
        b: float | NDArray[np.float64],
        a0: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        deg: int = 10,
    ) -> NDArray[np.float64]:
        a0 = reshape_ar_or_a0("a0", a0)
        return super().ls_integrate(func, a, b, *(a0, *args), deg=deg)

    @override
    def mean(
        self, a0: float | NDArray[np.float64], *args: float | NDArray[np.float64]
    ) -> np.float64 | NDArray[np.float64]:
        a0 = reshape_ar_or_a0("a0", a0)
        return super().mean(*(a0, *args))

    @override
    def median(
        self, a0: float | NDArray[np.float64], *args: float | NDArray[np.float64]
    ) -> np.float64 | NDArray[np.float64]:
        a0 = reshape_ar_or_a0("a0", a0)
        return super().median(*(a0, *args))

    @override
    def var(
        self, a0: float | NDArray[np.float64], *args: float | NDArray[np.float64]
    ) -> np.float64 | NDArray[np.float64]:
        a0 = reshape_ar_or_a0("a0", a0)
        return super().var(*(a0, *args))

    @override
    def moment(
        self, n: int, a0: float | NDArray[np.float64], *args: float | NDArray[np.float64]
    ) -> np.float64 | NDArray[np.float64]:
        a0 = reshape_ar_or_a0("a0", a0)
        return super().moment(n, *(a0, *args))

    @override
    def mrl(
        self, time: float | NDArray[np.float64], a0: float | NDArray[np.float64], *args: float | NDArray[np.float64]
    ) -> np.float64 | NDArray[np.float64]:
        a0 = reshape_ar_or_a0("a0", a0)
        return super().mrl(time, *(a0, *args))

    @override
    def ppf(
        self,
        probability: float | NDArray[np.float64],
        a0: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
    ) -> np.float64 | NDArray[np.float64]:
        a0 = reshape_ar_or_a0("a0", a0)
        return super().ppf(probability, *(a0, *args))


class FrozenAgeReplacementModel(
    FrozenParametricLifetimeModel[float | NDArray[np.float64], *tuple[float | NDArray[np.float64], ...]]
):
    r"""
    Frozen age replacement model.

    Parameters
    ----------
    model : AgeReplacementModel
        Any age replacement model.
    args_nb_assets : int
        Number of assets given in frozen arguments. It is automatically computed by ``freeze`` function.
    ar : float or np.ndarray
        Age of replacement values to be frozen.
    *args : float or np.ndarray
        Additional arguments needed by the model to be frozen.

    Attributes
    ----------
    unfrozen_model : AgeReplacementModel
        The unfrozen age replacement model.
    frozen_args : tuple of float or np.ndarray
        All the frozen arguments given and necessary to compute model functions.
    args_nb_assets : int
        Number of assets passed in frozen arguments. The data is mainly used to control numpy broadcasting and may not
        interest an user.

    Warnings
    --------
    The recommanded way to instanciate a frozen model is by using``freeze`` factory function.
    """

    unfrozen_model: AgeReplacementModel
    frozen_args: tuple[float | NDArray[np.float64], *tuple[float | NDArray[np.float64], ...]]

    @override
    def __init__(
        self,
        model: AgeReplacementModel,
        args_nb_assets: int,
        ar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
    ):
        super().__init__(model, args_nb_assets, *(ar, *args))

    @override
    def unfreeze(self) -> AgeReplacementModel:
        return super().unfreeze()

    @property
    def ar(self) -> float | NDArray[np.float64]:
        return self.frozen_args[0]

    @ar.setter
    def ar(self, value: float | NDArray[np.float64]) -> None:
        self.frozen_args = (value,) + self.frozen_args[1:]


class FrozenLeftTruncatedModel(
    FrozenParametricLifetimeModel[float | NDArray[np.float64], *tuple[float | NDArray[np.float64], ...]]
):
    r"""
    Frozen left truncated model.

    Parameters
    ----------
    model : LeftTruncatedModel
        Any left truncated model.
    args_nb_assets : int
        Number of assets given in frozen arguments. It is automatically computed by ``freeze`` function.
    a0 : float or np.ndarray
        Conditional age values to be frozen.
    *args : float or np.ndarray
        Additional arguments needed by the model to be frozen.


    Attributes
    ----------
    unfrozen_model : LeftTruncatedModel
        The unfrozen left truncated model.
    frozen_args : tuple of float or np.ndarray
        All the frozen arguments given and necessary to compute model functions.
    args_nb_assets : int
        Number of assets passed in frozen arguments. The data is mainly used to control numpy broadcasting and may not
        interest an user.

    Warnings
    --------
    The recommanded way to instanciate a frozen model is by using``freeze`` factory function.
    """

    unfrozen_model: LeftTruncatedModel
    frozen_args: tuple[float | NDArray[np.float64], *tuple[float | NDArray[np.float64], ...]]

    @override
    def __init__(
        self,
        model: LeftTruncatedModel,
        args_nb_assets: int,
        a0: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
    ):
        super().__init__(model, args_nb_assets, *(a0, *args))

    @override
    def unfreeze(self) -> LeftTruncatedModel:
        return super().unfreeze()

    @property
    def a0(self) -> float | NDArray[np.float64]:
        return self.frozen_args[0]

    @a0.setter
    def a0(self, value: float | NDArray[np.float64]) -> None:
        self.frozen_args = (value,) + self.frozen_args[1:]


A0_TIME_BASE_DOCSTRING = """
{name}.

Parameters
----------
time : float or np.ndarray
    Elapsed time value(s) at which to compute the function.
    If ndarray, allowed shapes are ``()``, ``(n_values,)`` or ``(n_assets, n_values)``.
a0 : float or np.ndarray
    Conditional age values. It represents ages reached by assets. If ndarray, shape can only be (m,) 
    as only one age per asset can be given
*args : float or np.ndarray
    Additional arguments needed by the model.

Returns
-------
np.float64 or np.ndarray
    Function values at each given time(s).
"""


A0_MOMENT_BASE_DOCSTRING = """
{name}.

Parameters
----------
a0 : float or np.ndarray
    Conditional age values. It represents ages reached by assets. If ndarray, shape can only be (m,) 
    as only one age per asset can be given
*args : float or np.ndarray
    Additional arguments needed by the model.

Returns
-------
np.float64
    {name} value.
"""

A0_PROBABILITY_BASE_DOCSTRING = """
{name}.

Parameters
----------
probability : float or np.ndarray
    Probability value(s) at which to compute the function.
    If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
a0 : float or np.ndarray
    Conditional age values. It represents ages reached by assets. If ndarray, shape can only be (m,) 
    as only one age per asset can be given
*args : float or np.ndarray
    Additional arguments needed by the model.

Returns
-------
np.float64 or np.ndarray
    Function values at each given probability value(s).
"""

LeftTruncatedModel.sf.__doc__ = A0_TIME_BASE_DOCSTRING.format(name="The survival function")
LeftTruncatedModel.hf.__doc__ = A0_TIME_BASE_DOCSTRING.format(name="The hazard function")
LeftTruncatedModel.chf.__doc__ = A0_TIME_BASE_DOCSTRING.format(name="The cumulative hazard function")
LeftTruncatedModel.pdf.__doc__ = A0_TIME_BASE_DOCSTRING.format(name="The probability density function")
LeftTruncatedModel.cdf.__doc__ = A0_TIME_BASE_DOCSTRING.format(name="The cumulative distribution function")
LeftTruncatedModel.mrl.__doc__ = A0_TIME_BASE_DOCSTRING.format(name="The mean residual life function")
LeftTruncatedModel.ppf.__doc__ = A0_PROBABILITY_BASE_DOCSTRING.format(name="The percent point function")
LeftTruncatedModel.ppf.__doc__ += f"""
Notes
-----
The ``ppf`` is the inverse of :py:meth:`~LeftTruncatedModel.cdf`.
"""
LeftTruncatedModel.isf.__doc__ = A0_PROBABILITY_BASE_DOCSTRING.format(name="Inverse survival function")

LeftTruncatedModel.rvs.__doc__ = """
Random variable sampling.

Parameters
----------
size : int, (int,) or (int, int)
    Size of the generated sample. If size is ``n`` or ``(n,)``, n samples are generated. If size is ``(m,n)``, a 
    2d array of samples is generated.
a0 : float or np.ndarray
    Conditional age values. It represents ages reached by assets. If ndarray, shape can only be (m,) 
    as only one age per asset can be given
*args : float or np.ndarray
    Additional arguments needed by the model. 
return_event : bool, default is False
    If True, returns event indicators along with the sample time values.
return_entry : bool, default is False
    If True, returns corresponding entry values of the sample time values.
seed : optional int, default is None
    Random seed used to fix random sampling.

Returns
-------
float, ndarray or tuple of float or ndarray
    The sample values. If either ``return_event`` or ``random_entry`` is True, returns a tuple containing
    the time values followed by event values, entry values or both.
"""

LeftTruncatedModel.ichf.__doc__ = """
Inverse cumulative hazard function.

Parameters
----------
cumulative_hazard_rate : float or np.ndarray
    Cumulative hazard rate value(s) at which to compute the function.
    If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

a0 : float or np.ndarray
    Conditional age values. It represents ages reached by assets. If ndarray, shape can only be (m,) 
    as only one age per asset can be given
*args : float or np.ndarray
    Additional arguments needed by the model.

Returns
-------
np.float64 or np.ndarray
    Function values at each given cumulative hazard rate(s).
"""

LeftTruncatedModel.plot.__doc__ = """
Provides access to plotting functionality for this distribution.
"""

LeftTruncatedModel.ls_integrate.__doc__ = """
Lebesgue-Stieltjes integration.

Parameters
----------
func : callable (in : 1 ndarray , out : 1 ndarray)
    The callable must have only one ndarray object as argument and one ndarray object as output
a : ndarray (maximum number of dimension is 2)
    Lower bound(s) of integration.
b : ndarray (maximum number of dimension is 2)
    Upper bound(s) of integration. If lower bound(s) is infinite, use np.inf as value.)
a0 : float or np.ndarray
    Conditional age values. It represents ages reached by assets. If ndarray, shape can only be (m,) 
    as only one age per asset can be given
*args : float or np.ndarray
    Additional arguments needed by the model.
deg : int, default 10
    Degree of the polynomials interpolation

Returns
-------
np.ndarray
    Lebesgue-Stieltjes integral of func from `a` to `b`.
"""

LeftTruncatedModel.moment.__doc__ = """
n-th order moment

Parameters
----------
n : order of the moment, at least 1.

Returns
-------
np.float64
    n-th order moment.
"""
LeftTruncatedModel.mean.__doc__ = A0_MOMENT_BASE_DOCSTRING.format(name="The mean")
LeftTruncatedModel.var.__doc__ = A0_MOMENT_BASE_DOCSTRING.format(name="The variance")
LeftTruncatedModel.median.__doc__ = A0_MOMENT_BASE_DOCSTRING.format(name="The median")

LeftTruncatedModel.ichf.__doc__ = """
Inverse cumulative hazard function.

Parameters
----------
cumulative_hazard_rate : float or np.ndarray
    Cumulative hazard rate value(s) at which to compute the function.
    If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

a0 : float or np.ndarray
    Conditional age values. It represents ages reached by assets. If ndarray, shape can only be (m,) 
    as only one age per asset can be given
*args : float or np.ndarray
    Additional arguments needed by the model.

Returns
-------
np.float64 or np.ndarray
    Function values at each given cumulative hazard rate(s).
"""


AR_TIME_BASE_DOCSTRING = """
{name}.

Parameters
----------
time : float or np.ndarray
    Elapsed time value(s) at which to compute the function.
    If ndarray, allowed shapes are ``()``, ``(n_values,)`` or ``(n_assets, n_values)``.
ar : float or np.ndarray
    Age of replacement values. If ndarray, shape can only be (m,) 
    as only one age of replacement per asset can be given
*args : float or np.ndarray
    Additional arguments needed by the model.

Returns
-------
np.float64 or np.ndarray
    Function values at each given time(s).
"""


AR_MOMENT_BASE_DOCSTRING = """
{name}.

Parameters
----------
ar : float or np.ndarray
    Age of replacement values. If ndarray, shape can only be (m,) 
    as only one age of replacement per asset can be given
*args : float or np.ndarray
    Additional arguments needed by the model.

Returns
-------
np.float64
    {name} value.
"""

AR_PROBABILITY_BASE_DOCSTRING = """
{name}.

Parameters
----------
probability : float or np.ndarray
    Probability value(s) at which to compute the function.
    If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
ar : float or np.ndarray
    Age of replacement values. If ndarray, shape can only be (m,) 
    as only one age of replacement per asset can be given
*args : float or np.ndarray
    Additional arguments needed by the model.

Returns
-------
np.float64 or np.ndarray
    Function values at each given probability value(s).
"""

AgeReplacementModel.sf.__doc__ = AR_TIME_BASE_DOCSTRING.format(name="The survival function")
AgeReplacementModel.hf.__doc__ = AR_TIME_BASE_DOCSTRING.format(name="The hazard function")
AgeReplacementModel.chf.__doc__ = AR_TIME_BASE_DOCSTRING.format(name="The cumulative hazard function")
AgeReplacementModel.pdf.__doc__ = AR_TIME_BASE_DOCSTRING.format(name="The probability density function")
AgeReplacementModel.cdf.__doc__ = AR_TIME_BASE_DOCSTRING.format(name="The cumulative distribution function")
AgeReplacementModel.mrl.__doc__ = AR_TIME_BASE_DOCSTRING.format(name="The mean residual life function")
AgeReplacementModel.ppf.__doc__ = AR_PROBABILITY_BASE_DOCSTRING.format(name="The percent point function")
AgeReplacementModel.isf.__doc__ = AR_PROBABILITY_BASE_DOCSTRING.format(name="Inverse survival function")

AgeReplacementModel.rvs.__doc__ = """
Random variable sampling.

Parameters
----------
size : int, (int,) or (int, int)
    Size of the generated sample. If size is ``n`` or ``(n,)``, n samples are generated. If size is ``(m,n)``, a 
    2d array of samples is generated.
ar : float or np.ndarray
    Age of replacement values. If ndarray, shape can only be (m,) 
    as only one age of replacement per asset can be given
*args : float or np.ndarray
    Additional arguments needed by the model. 
return_event : bool, default is False
    If True, returns event indicators along with the sample time values.
return_entry : bool, default is False
    If True, returns corresponding entry values of the sample time values.
seed : optional int, default is None
    Random seed used to fix random sampling.

Returns
-------
float, ndarray or tuple of float or ndarray
    The sample values. If either ``return_event`` or ``random_entry`` is True, returns a tuple containing
    the time values followed by event values, entry values or both.

Notes
-----
If ``return_entry`` is true, returned time values are not residual time. Otherwise, the times are residuals
"""

AgeReplacementModel.ichf.__doc__ = """
Inverse cumulative hazard function.

Parameters
----------
cumulative_hazard_rate : float or np.ndarray
    Cumulative hazard rate value(s) at which to compute the function.
    If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

ar : float or np.ndarray
    Age of replacement values. If ndarray, shape can only be (m,) 
    as only one age of replacement per asset can be given
*args : float or np.ndarray
    Additional arguments needed by the model.

Returns
-------
np.float64 or np.ndarray
    Function values at each given cumulative hazard rate(s).
"""


AgeReplacementModel.plot.__doc__ = """
Provides access to plotting functionality for this distribution.
"""

AgeReplacementModel.ls_integrate.__doc__ = """
Lebesgue-Stieltjes integration.

Parameters
----------
func : callable (in : 1 ndarray , out : 1 ndarray)
    The callable must have only one ndarray object as argument and one ndarray object as output
a : ndarray (maximum number of dimension is 2)
    Lower bound(s) of integration.
b : ndarray (maximum number of dimension is 2)
    Upper bound(s) of integration. If lower bound(s) is infinite, use np.inf as value.)
ar : float or np.ndarray
    Age of replacement values. If ndarray, shape can only be (m,) 
    as only one age of replacement per asset can be given
*args : float or np.ndarray
    Additional arguments needed by the model.
deg : int, default 10
    Degree of the polynomials interpolation

Returns
-------
np.ndarray
    Lebesgue-Stieltjes integral of func from `a` to `b`.
"""

AgeReplacementModel.moment.__doc__ = """
n-th order moment

Parameters
----------
n : order of the moment, at least 1.
ar : float or np.ndarray
    Age of replacement values. If ndarray, shape can only be (m,) 
    as only one age of replacement per asset can be given
*args : float or np.ndarray
    Additional arguments needed by the model.

Returns
-------
np.float64 or np.ndarray
    n-th order moment.
"""
AgeReplacementModel.mean.__doc__ = A0_MOMENT_BASE_DOCSTRING.format(name="The mean")
AgeReplacementModel.var.__doc__ = A0_MOMENT_BASE_DOCSTRING.format(name="The variance")
AgeReplacementModel.median.__doc__ = A0_MOMENT_BASE_DOCSTRING.format(name="The median")

AgeReplacementModel.ichf.__doc__ = """
Inverse cumulative hazard function.

Parameters
----------
cumulative_hazard_rate : float or np.ndarray
    Cumulative hazard rate value(s) at which to compute the function.
    If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
ar : float or np.ndarray
    Age of replacement values. If ndarray, shape can only be (m,) 
    as only one age of replacement per asset can be given
*args : float or np.ndarray
    Additional arguments needed by the model.

Returns
-------
np.float64 or np.ndarray
    Function values at each given cumulative hazard rate(s).
"""



from __future__ import annotations

from typing import (
    Any,
    Optional,
    Self,
    Sequence,
    Union,
)

import numpy as np
from numpy.typing import NDArray

from relife import ParametricModel
from relife.data import NHPPData
from relife.lifetime_model import LifetimeDistribution, LifetimeRegression


class NonHomogeneousPoissonProcess(ParametricModel):

    def __init__(self, baseline: LifetimeDistribution | LifetimeRegression):
        super().__init__()
        self.baseline = baseline

    @property
    def fitting_results(self):
        return self.baseline.fitting_results

    def intensity(self, time: float | NDArray[np.float64], *args: float | NDArray[np.float64]) -> NDArray[np.float64]:
        return self.baseline.hf(time, *args)

    def cumulative_intensity(
        self, time: float | NDArray[np.float64], *args: float | NDArray[np.float64]
    ) -> NDArray[np.float64]:
        return self.baseline.chf(time, *args)

    # def sample(
    #     self,
    #     size: int,
    #     tf: float,
    #     /,
    #     *args: float | NDArray[np.float64],
    #     t0: float = 0.0,
    #     maxsample: int = 1e5,
    #     seed: Optional[int] = None,
    # ):
    #
    # def failure_data_sample(
    #     self,
    #     size: int,
    #     tf: float,
    #     /,
    #     *args: float | NDArray[np.float64],
    #     t0: float = 0.0,
    #     maxsample: int = 1e5,
    #     seed: Optional[int] = None,
    # ) -> tuple[NDArray[np.float64], ...]:
    #     return self.freeze(*args).failure_data_sample(size, tf, t0=t0, maxsample=maxsample, seed=seed)

    # @property
    # def plot(self) -> PlotNHPP:
    #     return PlotNHPP(self)

    def fit(
        self,
        events_assets_ids: Union[Sequence[str], NDArray[np.int64]],
        events_ages: NDArray[np.float64],
        /,
        *args: float | NDArray[np.float64],
        assets_ids: Optional[Union[Sequence[str], NDArray[np.int64]]] = None,
        first_ages: Optional[NDArray[np.float64]] = None,
        last_ages: Optional[NDArray[np.float64]] = None,
        **kwargs: Any,
    ) -> Self:
        nhpp_data = NHPPData(
            events_assets_ids,
            events_ages,
            *args,
            assets_ids=assets_ids,
            first_ages=first_ages,
            last_ages=last_ages,
        )
        lifetime_data = nhpp_data.to_lifetime_data()
        self.baseline = self.baseline._fit_from_lifetime_data(lifetime_data)
        return self


class FrozenNonHomogeneousPoissonProcess(ParametricModel):
    def __init__(self, model: NonHomogeneousPoissonProcess, args_nb_assets: int, *args: float | NDArray[np.float64]):
        super().__init__()
        self.unfrozen_model = model
        self.frozen_args = args
        self.args_nb_assets = args_nb_assets

    def unfreeze(self) -> NonHomogeneousPoissonProcess:
        return self.unfrozen_model

    def intensity(self, time: float | NDArray[np.float64]) -> NDArray[np.float64]:
        return self.model.intensity(time, *self.frozen_args)

    def cumulative_intensity(self, time: float | NDArray[np.float64]) -> NDArray[np.float64]:
        return self.model.cumulative_intensity(time, *self.frozen_args)

    def sample_nhhp_data(self):
        pass

    def sample_count_data(self):
        pass

    # @property
    # def plot(self) -> PlotConstructor:
    #     return PlotNHPP(self)



"""
This module defines probability functions used in regression

Copyright (c) 2022, RTE (https://www.rte-france.com)
See AUTHORS.txt
SPDX-License-Identifier: Apache-2.0 (see LICENSE.txt)
"""

from __future__ import annotations

from typing import Literal, overload

import numpy as np
from numpy.typing import NDArray
from typing_extensions import override

from ._base import LifetimeRegression


def broadcast_time_covar(
    time: float | NDArray[np.float64],
    covar: float | NDArray[np.float64],
) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
    time = np.atleast_2d(np.asarray(time))  #  (m, n)
    covar = np.atleast_2d(np.asarray(covar))  #  (m, nb_coef)
    match (time.shape[0], covar.shape[0]):
        case (1, _):
            time = np.repeat(time, covar.shape[0], axis=0)
        case (_, 1):
            covar = np.repeat(covar, time.shape[0], axis=0)
        case (m1, m2) if m1 != m2:
            raise ValueError(f"Incompatible time and covar. time has {m1} nb_assets but covar has {m2} nb_assets")
    return time, covar


def broadcast_time_covar_shapes(
    time_shape: tuple[()] | tuple[int] | tuple[int, int],
    covar_shape: tuple[()] | tuple[int] | tuple[int, int],
) -> tuple[()] | tuple[int] | tuple[int, int]:
    # time_shape : (), (n,) or (m, n)
    # covar_shape : (), (nb_coef,) or (m, nb_coef)
    match [time_shape, covar_shape]:
        case [(), ()] | [(), (_,)]:
            return ()
        case [(), (m, _)]:
            return m, 1
        case [(n,), ()] | [(n,), (_,)]:
            return (n,)
        case [(n,), (m, _)] | [(m, n), ()] | [(m, n), (_,)]:
            return m, n
        case [(mt, n), (mc, _)] if mt != mc:
            if mt != 1 and mc != 1:
                raise ValueError(f"Invalid time and covar : time got {mt} nb assets but covar got {mc} nb assets")
            return max(mt, mc), n
        case [(mt, n), (mc, _)] if mt == mc:
            return mt, n
        case _:
            raise ValueError(f"Invalid time or covar shape. Got {time_shape} and {covar_shape}")


class ProportionalHazard(LifetimeRegression):
    # noinspection PyUnresolvedReferences
    r"""
    Proportional Hazard regression.

    The cumulative hazard function :math:`H` is linked to the multiplier
    function :math:`g` by the relation:

    .. math::

        H(t, x) = g(\beta, x) H_0(t) = e^{\beta \cdot x} H_0(t)

    where :math:`x` is a vector of covariates, :math:`\beta` is the coefficient
    vector of the effect of covariates, :math:`H_0` is the baseline cumulative
    hazard function [1]_.

    |

    Parameters
    ----------
    baseline : FittableParametricLifetimeModel
        Any lifetime model that can be fitted.
    coefficients : tuple of floats (values can be None), default is (None,)
        Coefficients values of the covariate effects.

    Attributes
    ----------
    baseline : FittableParametricLifetimeModel
        The regression baseline model (lifetime model).
    covar_effect : CovarEffect
        The regression covariate effect.
    fitting_results : FittingResults, default is None
        An object containing fitting results (AIC, BIC, etc.). If the model is not fitted, the value is None.
    coefficients
    nb_params
    params
    params_names
    plot


    References
    ----------
    .. [1] Sun, J. (2006). The statistical analysis of interval-censored failure
        time data (Vol. 3, No. 1). New York: springer.

    See Also
    --------
    regression.AFT : Accelerated Failure Time regression.

    """

    def hf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
    ) -> np.float64 | NDArray[np.float64]:
        return self.covar_effect.g(covar) * self.baseline.hf(time, *args)

    def chf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
    ) -> np.float64 | NDArray[np.float64]:
        # (m,)
        return self.covar_effect.g(covar) * self.baseline.chf(time, *args)

    @override
    def ichf(
        self,
        cumulative_hazard_rate: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
    ) -> np.float64 | NDArray[np.float64]:
        return self.baseline.ichf(cumulative_hazard_rate / self.covar_effect.g(covar), *args)

    def dhf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
    ) -> np.float64 | NDArray[np.float64]:
        return self.covar_effect.g(covar) * self.baseline.dhf(time, *args)

    @overload
    def jac_hf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        asarray: Literal[False] = False,
    ) -> tuple[np.float64 | NDArray[np.float64], ...]: ...

    @overload
    def jac_hf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        asarray: Literal[True] = True,
    ) -> np.float64 | NDArray[np.float64]: ...

    def jac_hf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        asarray: bool = False,
    ) -> np.float64 | NDArray[np.float64] | tuple[np.float64 | NDArray[np.float64], ...]:

        time = np.asarray(time)  # (), (n,) or (m, n)
        covar = np.asarray(covar)  # (), (nb_coef,) or (m, nb_coef)
        out_shape = broadcast_time_covar_shapes(time.shape, covar.shape)  # (), (n,) or (m, n)
        time, covar = broadcast_time_covar(time, covar)  # (m, n) and (m, nb_coef)

        g = self.covar_effect.g(covar)  # (m, 1)
        jac_g = self.covar_effect.jac_g(covar, asarray=True)  # (nb_coef, m, 1)

        baseline_hf = self.baseline.hf(time, *args)  # (m, n)
        # p == baseline.nb_params
        baseline_jac_hf = self.baseline.jac_hf(time, *args, asarray=True)  # (p, m, n)
        jac_g = np.repeat(jac_g, baseline_hf.shape[-1], axis=-1)  # (nb_coef, m, n) necessary to concatenate

        jac = np.concatenate(
            (
                baseline_hf * jac_g,  #  (nb_coef, m, n)
                g * baseline_jac_hf,  # (p, m, n)
            ),
            axis=0,
        )  # (p + nb_coef, m, n)

        jac = jac.reshape((self.nb_params,) + out_shape)
        if not asarray:
            return np.unstack(jac)
        return jac

    @overload
    def jac_chf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        asarray: Literal[False] = False,
    ) -> tuple[np.float64 | NDArray[np.float64], ...]: ...

    @overload
    def jac_chf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        asarray: Literal[True] = True,
    ) -> np.float64 | NDArray[np.float64]: ...

    def jac_chf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        asarray: bool = False,
    ) -> np.float64 | NDArray[np.float64] | tuple[np.float64 | NDArray[np.float64], ...]:

        time = np.asarray(time)  # (), (n,) or (m, n)
        covar = np.asarray(covar)  # (), (nb_coef,) or (m, nb_coef)
        out_shape = broadcast_time_covar_shapes(time.shape, covar.shape)  # (), (n,) or (m, n)
        time, covar = broadcast_time_covar(time, covar)  # (m, n) and (m, nb_coef)

        g = self.covar_effect.g(covar)  # (m, 1)
        jac_g = self.covar_effect.jac_g(covar, asarray=True)  # (nb_coef, m, 1)
        baseline_chf = self.baseline.chf(time, *args)  # (m, n)
        #  p == baseline.nb_params
        baseline_jac_chf = self.baseline.jac_chf(time, *args, asarray=True)  # (p, m, n)
        jac_g = np.repeat(jac_g, baseline_chf.shape[-1], axis=-1)  # (nb_coef, m, n) necessary to concatenate

        jac = np.concatenate(
            (
                baseline_chf * jac_g,  #  (nb_coef, m, n)
                g * baseline_jac_chf,  # (p, m, n)
            ),
            axis=0,
        )  # (p + nb_coef, m, n)

        jac = jac.reshape((self.nb_params,) + out_shape)
        if not asarray:
            return np.unstack(jac)
        return jac

    @override
    def moment(
        self, n: int, covar: float | NDArray[np.float64], *args: float | NDArray[np.float64]
    ) -> np.float64 | NDArray[np.float64]:
        return super().moment(n, *(covar, *args))


class AcceleratedFailureTime(LifetimeRegression):
    # noinspection PyUnresolvedReferences
    r"""
    Accelerated failure time regression.

    The cumulative hazard function :math:`H` is linked to the multiplier
    function :math:`g` by the relation:

    .. math::

        H(t, x) = H_0\left(\dfrac{t}{g(\beta, x)}\right) = H_0(t e^{- \beta
        \cdot x})

    where :math:`x` is a vector of covariates, :math:`\beta` is the coefficient
    vector of the effect of covariates, :math:`H_0` is the baseline cumulative
    hazard function [1]_.

    |

    Parameters
    ----------
    baseline : FittableParametricLifetimeModel
        Any lifetime model that can be fitted.
    coefficients : tuple of floats (values can be None), default is (None,)
        Coefficients values of the covariate effects.

    Attributes
    ----------
    baseline : FittableParametricLifetimeModel
        The regression baseline model (lifetime model).
    covar_effect : CovarEffect
        The regression covariate effect.
    fitting_results : FittingResults, default is None
        An object containing fitting results (AIC, BIC, etc.). If the model is not fitted, the value is None.
    coefficients
    nb_params
    params
    params_names
    plot


    References
    ----------
    .. [1] Kalbfleisch, J. D., & Prentice, R. L. (2011). The statistical
        analysis of failure time data. John Wiley & Sons.

    See Also
    --------
    regression.ProportionalHazard : proportional hazard regression
    """

    def hf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
    ) -> np.float64 | NDArray[np.float64]:
        t0 = time / self.covar_effect.g(covar)
        return self.baseline.hf(t0, *args) / self.covar_effect.g(covar)

    def chf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
    ) -> np.float64 | NDArray[np.float64]:
        t0 = time / self.covar_effect.g(covar)
        return self.baseline.chf(t0, *args)

    @override
    def ichf(
        self,
        cumulative_hazard_rate: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
    ) -> np.float64 | NDArray[np.float64]:
        return self.covar_effect.g(covar) * self.baseline.ichf(cumulative_hazard_rate, *args)

    def dhf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
    ) -> np.float64 | NDArray[np.float64]:
        t0 = time / self.covar_effect.g(covar)
        return self.baseline.dhf(t0, *args) / self.covar_effect.g(covar) ** 2

    @overload
    def jac_hf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        asarray: Literal[False] = False,
    ) -> tuple[np.float64 | NDArray[np.float64], ...]: ...

    @overload
    def jac_hf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        asarray: Literal[True] = True,
    ) -> np.float64 | NDArray[np.float64]: ...

    def jac_hf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        asarray: bool = False,
    ) -> np.float64 | NDArray[np.float64] | tuple[np.float64 | NDArray[np.float64], ...]:

        time = np.asarray(time)  # (), (n,) or (m, n)
        covar = np.asarray(covar)  # (), (nb_coef,) or (m, nb_coef)
        out_shape = broadcast_time_covar_shapes(time.shape, covar.shape)  # (), (n,) or (m, n)
        time, covar = broadcast_time_covar(time, covar)  # (m, n) and (m, nb_coef)

        g = self.covar_effect.g(covar)  # (m, 1)
        jac_g = self.covar_effect.jac_g(covar, asarray=True)  # (nb_coef, m, 1)
        t0 = time / g  # (m, n)
        # p == baseline.nb_params
        baseline_jac_hf_t0 = self.baseline.jac_hf(t0, *args, asarray=True)  # (p, m, n)
        baseline_hf_t0 = self.baseline.hf(t0, *args)  # (m, n)
        baseline_dhf_t0 = self.baseline.dhf(t0, *args)  # (m, n)
        jac_g = np.repeat(jac_g, baseline_hf_t0.shape[-1], axis=-1)  # (nb_coef, m, n)

        jac = np.concatenate(
            (
                -jac_g / g**2 * (baseline_hf_t0 + t0 * baseline_dhf_t0),  # (nb_coef, m, n) necessary to concatenate
                baseline_jac_hf_t0 / g,  # (p, m, n)
            ),
            axis=0,
        )  # (p + nb_coef, m, n)

        jac = jac.reshape((self.nb_params,) + out_shape)
        if not asarray:
            return np.unstack(jac)
        return jac

    @overload
    def jac_chf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        asarray: Literal[False] = False,
    ) -> tuple[np.float64 | NDArray[np.float64], ...]: ...

    @overload
    def jac_chf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        asarray: Literal[True] = True,
    ) -> np.float64 | NDArray[np.float64]: ...

    def jac_chf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        asarray: bool = False,
    ) -> np.float64 | NDArray[np.float64] | tuple[np.float64 | NDArray[np.float64], ...]:

        time = np.asarray(time)  # (), (n,) or (m, n)
        covar = np.asarray(covar)  # (), (nb_coef,) or (m, nb_coef)
        out_shape = broadcast_time_covar_shapes(time.shape, covar.shape)  # (), (n,) or (m, n)
        time, covar = broadcast_time_covar(time, covar)  # (m, n) and (m, nb_coef)

        g = self.covar_effect.g(covar)  # (m, 1)
        jac_g = self.covar_effect.jac_g(covar, asarray=True)  # (nb_coef, m, 1)
        t0 = time / g  #  (m, n)
        # p == baseline.nb_params
        baseline_jac_chf_t0 = self.baseline.jac_chf(t0, *args, asarray=True)  # (p, m, n)
        baseline_hf_t0 = self.baseline.hf(t0, *args)  #  (m, n)
        jac_g = np.repeat(jac_g, baseline_hf_t0.shape[-1], axis=-1)  # (nb_coef, m, n) necessary to concatenate

        jac = np.concatenate(
            (
                -jac_g / g * t0 * baseline_hf_t0,  #  (nb_coef, m, n)
                baseline_jac_chf_t0,  # (p, m, n)
            ),
            axis=0,
        )  # (p + nb_coef, m, n)

        jac = jac.reshape((self.nb_params,) + out_shape)
        if not asarray:
            return np.unstack(jac)
        return jac

    @override
    def moment(
        self, n: int, covar: float | NDArray[np.float64], *args: float | NDArray[np.float64]
    ) -> np.float64 | NDArray[np.float64]:
        return super().moment(n, *(covar, *args))


TIME_BASE_DOCSTRING = """
{name}.

Parameters
----------
time : float or np.ndarray
    Elapsed time value(s) at which to compute the function.
    If ndarray, allowed shapes are ``()``, ``(n_values,)`` or ``(n_assets, n_values)``.
covar : float or np.ndarray
    Covariates values. float can only be valid if the regression has one coefficients. 
    Otherwise it must be a ndarray of shape (nb_coef,) or (m, nb_coef)
*args : float or np.ndarray
    Additional arguments needed by the model.

Returns
-------
np.float64 or np.ndarray
    Function values at each given time(s).
"""

JAC_BASE_DOCSTRING = """
{name}.

Parameters
----------
time : float or np.ndarray
    Elapsed time value(s) at which to compute the function.
    If ndarray, allowed shapes are ``()``, ``(n_values,)`` or ``(n_assets, n_values)``.
covar : float or np.ndarray
    Covariates values. float can only be valid if the regression has one coefficients. 
    Otherwise it must be a ndarray of shape (nb_coef,) or (m, nb_coef)
*args : float or np.ndarray
    Additional arguments needed by the model.
asarray : bool, default is False

Returns
-------
np.float64, np.ndarray or tuple of np.float64 or np.ndarray
    The derivatives with respect to each parameter. If ``asarray`` is False, the function returns a tuple containing
    the same number of elements as parameters. If ``asarray`` is True, the function returns an ndarray
    whose first dimension equals the number of parameters. This output is equivalent to applying ``np.stack`` on the output
    tuple when ``asarray`` is False.
"""

MOMENT_BASE_DOCSTRING = """
{name}.

Parameters
----------
covar : float or np.ndarray
    Covariates values. float can only be valid if the regression has one coefficients. 
    Otherwise it must be a ndarray of shape ``(nb_coef,)`` or ``(m, nb_coef)``.
*args : float or np.ndarray
    Additional arguments needed by the model.

Returns
-------
np.float64
    {name} value.
"""

ICHF_DOCSTRING = """
Inverse cumulative hazard function.

Parameters
----------
cumulative_hazard_rate : float or np.ndarray
    Cumulative hazard rate value(s) at which to compute the function.
    If ndarray, allowed shapes are ``()``, ``(n_values,)`` or ``(n_assets, n_values)``.
covar : float or np.ndarray
    Covariates values. float can only be valid if the regression has one coefficients. 
    Otherwise it must be a ndarray of shape ``(nb_coef,)`` or ``(m, nb_coef)``.
*args : float or np.ndarray
    Additional arguments needed by the model.

Returns
-------
np.float64 or np.ndarray
    Function values at each given cumulative hazard rate(s).
"""


PROBABILITY_BASE_DOCSTRING = """
{name}.

Parameters
----------
probability : float or np.ndarray
    Probability value(s) at which to compute the function.
    If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
covar : float or np.ndarray
    Covariates values. float can only be valid if the regression has one coefficients. 
    Otherwise it must be a ndarray of shape ``(nb_coef,)`` or ``(m, nb_coef)``.
*args : float or np.ndarray
    Additional arguments needed by the model.

Returns
-------
np.float64 or np.ndarray
    Function values at each given probability value(s).
"""


for class_obj in (AcceleratedFailureTime, ProportionalHazard):
    class_obj.sf.__doc__ = TIME_BASE_DOCSTRING.format(name="The survival function")
    class_obj.hf.__doc__ = TIME_BASE_DOCSTRING.format(name="The hazard function")
    class_obj.chf.__doc__ = TIME_BASE_DOCSTRING.format(name="The cumulative hazard function")
    class_obj.pdf.__doc__ = TIME_BASE_DOCSTRING.format(name="The probability density function")
    class_obj.cdf.__doc__ = TIME_BASE_DOCSTRING.format(name="The cumulative distribution function")
    class_obj.mrl.__doc__ = TIME_BASE_DOCSTRING.format(name="The mean residual life function")
    class_obj.dhf.__doc__ = TIME_BASE_DOCSTRING.format(name="The derivative of the hazard function")
    class_obj.jac_hf.__doc__ = JAC_BASE_DOCSTRING.format(name="The jacobian of the hazard function")
    class_obj.jac_chf.__doc__ = JAC_BASE_DOCSTRING.format(name="The jacobian of the cumulative hazard function")
    class_obj.jac_sf.__doc__ = JAC_BASE_DOCSTRING.format(name="The jacobian of the survival function")
    class_obj.jac_pdf.__doc__ = JAC_BASE_DOCSTRING.format(name="The jacobian of the probability density function")
    class_obj.jac_cdf.__doc__ = JAC_BASE_DOCSTRING.format(name="The jacobian of the cumulative distribution function")

    class_obj.ppf.__doc__ = PROBABILITY_BASE_DOCSTRING.format(name="The percent point function")
    class_obj.ppf.__doc__ += f"""
    Notes
    -----
    The ``ppf`` is the inverse of :py:meth:`~{class_obj}.cdf`.
    """
    class_obj.isf.__doc__ = PROBABILITY_BASE_DOCSTRING.format(name="Inverse survival function")

    class_obj.rvs.__doc__ = """
    Random variable sampling.

    Parameters
    ----------
    size : int, (int,) or (int, int)
        Size of the generated sample. If size is ``n`` or ``(n,)``, n samples are generated. If size is ``(m,n)``, a 
        2d array of samples is generated. 
    return_event : bool, default is False
        If True, returns event indicators along with the sample time values.
    random_entry : bool, default is False
        If True, returns corresponding entry values of the sample time values.
    seed : optional int, default is None
        Random seed used to fix random sampling.

    Returns
    -------
    float, ndarray or tuple of float or ndarray
        The sample values. If either ``return_event`` or ``random_entry`` is True, returns a tuple containing
        the time values followed by event values, entry values or both.
    """

    class_obj.plot.__doc__ = """
    Provides access to plotting functionality for this distribution.
    """

    class_obj.ls_integrate.__doc__ = """
    Lebesgue-Stieltjes integration.

    Parameters
    ----------
    func : callable (in : 1 ndarray , out : 1 ndarray)
        The callable must have only one ndarray object as argument and one ndarray object as output
    a : ndarray (maximum number of dimension is 2)
        Lower bound(s) of integration.
    b : ndarray (maximum number of dimension is 2)
        Upper bound(s) of integration. If lower bound(s) is infinite, use np.inf as value.)
    covar : float or np.ndarray
        Covariates values. float can only be valid if the regression has one coefficients. 
        Otherwise it must be a ndarray of shape ``(nb_coef,)`` or ``(m, nb_coef)``.
    *args : float or np.ndarray
        Additional arguments needed by the model.
    deg : int, default 10
        Degree of the polynomials interpolation

    Returns
    -------
    np.ndarray
        Lebesgue-Stieltjes integral of func from `a` to `b`.
    """

    class_obj.moment.__doc__ = """
    n-th order moment

    Parameters
    ----------
    n : order of the moment, at least 1.

    Returns
    -------
    np.float64
        n-th order moment.
    """
    class_obj.mean.__doc__ = MOMENT_BASE_DOCSTRING.format(name="The mean")
    class_obj.var.__doc__ = MOMENT_BASE_DOCSTRING.format(name="The variance")
    class_obj.median.__doc__ = MOMENT_BASE_DOCSTRING.format(name="The median")

    class_obj.ichf.__doc__ = """
    Inverse cumulative hazard function.

    Parameters
    ----------
    cumulative_hazard_rate : float or np.ndarray
        Cumulative hazard rate value(s) at which to compute the function.
        If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        
    covar : float or np.ndarray
        Covariates values. float can only be valid if the regression has one coefficients. 
        Otherwise it must be a ndarray of shape ``(nb_coef,)`` or ``(m, nb_coef)``.
    *args : float or np.ndarray
        Additional arguments needed by the model.

    Returns
    -------
    np.float64 or np.ndarray
        Function values at each given cumulative hazard rate(s).
    """

    class_obj.fit.__doc__ = """
    Estimation of parameters.

    Parameters
    ----------
    time : ndarray (1d or 2d)
        Observed lifetime values.
    covar : float or np.ndarray
        Covariates values. float can only be valid if the regression has one coefficients. 
        Otherwise it must be a ndarray of shape ``(nb_coef,)`` or ``(m, nb_coef)``.
    *args : float or np.ndarray
        Additional arguments needed by the model.
    event : ndarray of boolean values (1d), default is None
        Boolean indicators tagging lifetime values as right censored or complete.
    entry : ndarray of float (1d), default is None
        Left truncations applied to lifetime values.
    departure : ndarray of float (1d), default is None
        Right truncations applied to lifetime values.
    **kwargs
        Extra arguments used by `scipy.minimize`. Default values are:
            - `method` : `"L-BFGS-B"`
            - `contraints` : `()`
            - `tol` : `None`
            - `callback` : `None`
            - `options` : `None`
            - `bounds` : `self.params_bounds`
            - `x0` : `self.init_params`

    Returns
    -------
    Self
        The current object with the estimated parameters setted inplace.

    Notes
    -----
    Supported lifetime observations format is either 1d-array or 2d-array. 2d-array is more advanced
    format that allows to pass other information as left-censored or interval-censored values. In this case,
    `event` is not needed as 2d-array encodes right-censored values by itself.
    """
